/*****************************************************************************
 * Copyright (c) 2014-2026 OpenRCT2 developers
 *
 * For a complete list of all authors, please refer to contributors.md
 * Interested in contributing? Visit https://github.com/OpenRCT2/OpenRCT2
 *
 * OpenRCT2 is licensed under the GNU General Public License version 3.
 *****************************************************************************/

#pragma once

#include "../GameAction.hpp"

namespace OpenRCT2::GameActions
{
    enum class ModifyGroupType : uint8_t
    {
        addGroup,
        removeGroup,
        setPermissions,
        setName,
        setDefault,
        count
    };

    enum class PermissionState : uint8_t
    {
        toggle,
        setAll,
        clearAll,
        count
    };

    class NetworkModifyGroupAction final : public GameActionBase<GameCommand::ModifyGroups>
    {
    private:
        ModifyGroupType _type{ ModifyGroupType::count };
        uint8_t _groupId{ std::numeric_limits<uint8_t>::max() };
        std::string _name;
        uint32_t _permissionIndex{ std::numeric_limits<uint32_t>::max() };
        PermissionState _permissionState{ PermissionState::count };

    public:
        NetworkModifyGroupAction() = default;
        NetworkModifyGroupAction(
            ModifyGroupType type, uint8_t groupId = std::numeric_limits<uint8_t>::max(), std::string name = "",
            uint32_t permissionIndex = 0, PermissionState permissionState = PermissionState::count);

        void AcceptParameters(GameActionParameterVisitor&) final;

        uint16_t GetActionFlags() const override;

        void Serialise(DataSerialiser& stream) override;
        Result Query(GameState_t& gameState, Park::ParkData& park) const override;
        Result Execute(GameState_t& gameState, Park::ParkData& park) const override;
    };
} // namespace OpenRCT2::GameActions
