/*****************************************************************************
 * Copyright (c) 2014-2026 OpenRCT2 developers
 *
 * For a complete list of all authors, please refer to contributors.md
 * Interested in contributing? Visit https://github.com/OpenRCT2/OpenRCT2
 *
 * OpenRCT2 is licensed under the GNU General Public License version 3.
 *****************************************************************************/

#pragma once

#include <cstdint>

namespace OpenRCT2::Drawing
{
    enum class Colour : uint8_t;
}

struct VehicleColour
{
    OpenRCT2::Drawing::Colour Body;
    OpenRCT2::Drawing::Colour Trim;
    OpenRCT2::Drawing::Colour Tertiary;
};
