#pragma once

#include "NetworkBase.h"

#ifndef DISABLE_NETWORK

namespace OpenRCT2::Network
{
    class NetworkClient final : public NetworkBase
    {
    public:
    };
} // namespace OpenRCT2::Network

#endif // DISABLE_NETWORK
