/*****************************************************************************
 * Copyright (c) 2014-2026 OpenRCT2 developers
 *
 * For a complete list of all authors, please refer to contributors.md
 * Interested in contributing? Visit https://github.com/OpenRCT2/OpenRCT2
 *
 * OpenRCT2 is licensed under the GNU General Public License version 3.
 *****************************************************************************/

#pragma once

#include "../GameAction.hpp"

namespace OpenRCT2::GameActions
{
    enum class TileModifyType : uint8_t
    {
        anyRemove,
        anySwap,
        anyToggleInvisilibity,
        anyRotate,
        anyPaste,
        anySort,
        anyBaseHeightOffset,
        surfaceShowParkFences,
        surfaceToggleCorner,
        surfaceToggleDiagonal,
        pathSetSlope,
        pathSetBroken,
        pathToggleEdge,
        pathSetJunctionRailings,
        entranceMakeUsable,
        wallSetSlope,
        wallSetAnimationFrame,
        trackBaseHeightOffset,
        trackSetChain,
        trackSetChainBlock,
        trackSetBrake,
        trackSetIndestructible,
        scenerySetQuarterLocation,
        scenerySetQuarterCollision,
        bannerToggleBlockingEdge,
        wallSetAnimationIsBackwards,
        count,
    };

    class TileModifyAction final : public GameActionBase<GameCommand::ModifyTile>
    {
    private:
        CoordsXY _loc;
        TileModifyType _setting{};
        uint32_t _value1{};
        uint32_t _value2{};
        TileElement _pasteElement{};
        Banner _pasteBanner{};

    public:
        TileModifyAction() = default;
        TileModifyAction(
            CoordsXY loc, TileModifyType setting, uint32_t value1 = 0, uint32_t value2 = 0, TileElement pasteElement = {},
            Banner _pasteBanner = {});

        void AcceptParameters(GameActionParameterVisitor&) final;

        uint16_t GetActionFlags() const override;

        void Serialise(DataSerialiser& stream) override;
        Result Query(GameState_t& gameState, Park::ParkData& park) const override;
        Result Execute(GameState_t& gameState, Park::ParkData& park) const override;

    private:
        Result QueryExecute(bool isExecuting) const;
    };
} // namespace OpenRCT2::GameActions
